@@ -18,114 +18,12 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with the Gnome Library; see the file COPYING.LIB.  If not,
- * write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
- * Boston, MA 02110-1301, USA.
+ * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+ * Boston, MA 02111-1307, USA.
  *
  * Author: Alexander Larsson <alexl@redhat.com>
  */
 
-/**
- * SECTION:gnome-desktop-thumbnail
- * @short_description: Generates and looks up thumbnails of files and
- * directories
- * @stability: Unstable
- * @include: libgnome-desktop/gnome-desktop-thumbnail.h
- *
- * #GnomeDesktopThumbnailFactory allows generation and loading of thumbnails for
- * local and remote files and directories. It uses a collection of programs
- * called <firstterm>thumbnailers</firstterm>, each one generating thumbnails
- * for a specific set of content-types of files. For example,
- * <application>totem-video-thumbnailer</application> generates thumbnails for
- * video files using GStreamer; <application>evince-thumbnailer</application>
- * generates thumbnails for PDFs and other document files. If no specific
- * thumbnailer exists for a file, or if the thumbnailer fails, gdk-pixbuf is
- * used as a fallback.
- *
- * To generate a thumbnail, an appropriate thumbnailer program is selected then
- * executed, passing it the URI of the file to thumbnail, plus a path to write
- * the thumbnail image to. If thumbnailing succeeds, the thumbnailer should have
- * written the image to disk before terminating; but if thumbnailing fails, no
- * image should be written, and the thumbnailer should return a non-zero exit
- * status. #GnomeDesktopThumbnailFactory will then fall back to using gdk-pixbuf
- * to generate a thumbnail, if possible.
- *
- * Thumbnailers are chosen by examining a series of
- * <filename>.thumbnailer</filename> files in
- * <filename><replaceable>$PREFIX</replaceable>/share/thumbnailers</filename>.
- * Each is in a simple key-file format:
- * <informalexample><programlisting>
- * [Thumbnailer Entry]
- * TryExec=evince-thumbnailer
- * Exec=evince-thumbnailer -s %s %u %o
- * MimeType=application/pdf;application/x-bzpdf;application/x-gzpdf;
- * </programlisting></informalexample>
- *
- * The <filename>.thumbnailer</filename> format supports three keys:
- * <variablelist>
- * <varlistentry><term><code>TryExec</code></term><listitem><para>
- * Optional. The name of the the thumbnailer program in the current
- * <envar>$PATH</envar>. This allows the calling code to quickly check whether
- * the thumbnailer exists before trying to execute it.
- * </para></listitem></varlistentry>
- * <varlistentry><term><code>Exec</code></term><listitem><para>
- * Required. The command to execute the thumbnailer. It supports a few different
- * parameters which are replaced before calling the thumbnailer:
- * <replaceable>%u</replaceable> is the URI of the file being thumbnailed;
- * <replaceable>%i</replaceable> is its path; <replaceable>%o</replaceable>
- * is the path of the image file to be written to;
- * <replaceable>%s</replaceable> is the maximum desired size of the thumbnail
- * image (the maximum width or height, in pixels); and
- * <replaceable>%%</replaceable> is a literal percent character.
- * </para></listitem></varlistentry>
- * <varlistentry><term><code>MimeType</code></term><listitem><para>
- * Required. A semicolon-separated list of MIME types which the thumbnailer
- * supports generating thumbnails for.
- * </para></listitem></varlistentry>
- * </variablelist>
- *
- * So in the example <filename>.thumbnailer</filename> file above, the command
- * passes the requested thumbnail size, then the input file’s URI, then the
- * path for the output image file to
- * <application>evince-thumbnailer</application>.
- *
- * The code to examine and call a thumbnailer is contained in
- * #GnomeDesktopThumbnailFactory, which handles looking up the right thumbnailer
- * script, building and executing the command for it, and loading the resulting
- * thumbnail image into a #GdkPixbuf.
- *
- * Thumbnail caching is also supported by #GnomeDesktopThumbnailFactory. When
- * calling a thumbnailer, the path passed for the output image file is in
- * <filename><envar>$XDG_CACHE_HOME</envar>/thumbnails/
- * <replaceable>$SIZE</replaceable>/</filename>. The cached image file is given
- * a (probably) unique filename, generated by hashing the original file’s URI,
- * so the thumbnail can be looked up in future. #GnomeDesktopThumbnailFactory
- * supports two sizes of thumbnails: %GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL and
- * %GNOME_DESKTOP_THUMBNAIL_SIZE_LARGE. Normal thumbnails are up to 128×128
- * pixels, whereas large thumbnails are up to 256×256 pixels. Thumbnails which
- * are larger than this are scaled down before being cached, and non-square
- * thumbnails are scaled so their largest dimension is at most 128 or 256
- * pixels.
- *
- * #GnomeDesktopThumbnailFactory also handles failed thumbnails. If a
- * thumbnailer can’t generate a thumbnail for a file (e.g. because the file is
- * corrupt or because the right video codecs aren’t available), it returns a
- * non-zero exit status. The thumbnail factory then writes an entry to
- * <filename><envar>$XDG_CACHE_HOME</envar>/thumbnails/fail/
- * gnome-thumbnail-factory/</filename> which is named after the hash of the
- * input file URI (just like a successful cached thumbnail). For future queries
- * for thumbnails for that file, #GnomeDesktopThumbnailFactory can immediately
- * return an error after looking up the fail entry.
- *
- * If a file changes content, #GnomeDesktopThumbnailFactory will generate a new
- * thumbnail because each cached image has associated metadata (stored as PNG
- * tEXt keys) storing the full URI of the thumbnailed file (to check for hash
- * collisions) and its last modification time at the point of thumbnailing. If
- * the stored modification time doesn’t match the file’s current one, a new
- * thumbnail is generated.
- *
- * Since: 2.2
- */
-
 #include <config.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -148,17 +46,10 @@
 
 #define SECONDS_BETWEEN_STATS 10
 
-static void
-thumbnailers_directory_changed (GFileMonitor                 *monitor,
-                                GFile                        *file,
-                                GFile                        *other_file,
-                                GFileMonitorEvent             event_type,
-                                GnomeDesktopThumbnailFactory *factory);
-
 struct _GnomeDesktopThumbnailFactoryPrivate {
   GnomeDesktopThumbnailSize size;
 
-  GMutex lock;
+  GMutex *lock;
 
   GList *thumbnailers;
   GHashTable *mime_types_map;
@@ -172,6 +63,9 @@ struct _GnomeDesktopThumbnailFactoryPrivate {
 
 static const char *appname = "gnome-thumbnail-factory";
 
+static void gnome_desktop_thumbnail_factory_init          (GnomeDesktopThumbnailFactory      *factory);
+static void gnome_desktop_thumbnail_factory_class_init    (GnomeDesktopThumbnailFactoryClass *class);
+
 G_DEFINE_TYPE (GnomeDesktopThumbnailFactory,
 	       gnome_desktop_thumbnail_factory,
 	       G_TYPE_OBJECT)
@@ -403,39 +297,6 @@ size_prepared_cb (GdkPixbufLoader *loader,
   gdk_pixbuf_loader_set_size (loader, width, height);
 }
 
-static GdkPixbufLoader *
-create_loader (GFile        *file,
-               const guchar *data,
-               gsize         size)
-{
-  GdkPixbufLoader *loader;
-  GError *error = NULL;
-  char *mime_type;
-  char *filename;
-
-  loader = NULL;
-
-  /* need to specify the type here because the gdk_pixbuf_loader_write
-     doesn't have access to the filename in order to correct detect
-     the image type. */
-  filename = g_file_get_basename (file);
-  mime_type = g_content_type_guess (filename, data, size, NULL);
-  g_free (filename);
-
-  if (mime_type != NULL) {
-    loader = gdk_pixbuf_loader_new_with_mime_type (mime_type, &error);
-  }
-
-  if (loader == NULL) {
-    g_warning ("Unable to create loader for mime type %s: %s", mime_type, error->message);
-    g_clear_error (&error);
-    loader = gdk_pixbuf_loader_new ();
-  }
-  g_free (mime_type);
-
-  return loader;
-}
-
 static GdkPixbuf *
 _gdk_pixbuf_new_from_uri_at_scale (const char *uri,
 				   gint        width,
@@ -443,9 +304,9 @@ _gdk_pixbuf_new_from_uri_at_scale (const char *uri,
 				   gboolean    preserve_aspect_ratio)
 {
     gboolean result;
-    guchar buffer[LOAD_BUFFER_SIZE];
-    gssize bytes_read;
-    GdkPixbufLoader *loader = NULL;
+    char buffer[LOAD_BUFFER_SIZE];
+    gsize bytes_read;
+    GdkPixbufLoader *loader;
     GdkPixbuf *pixbuf;	
     GdkPixbufAnimation *animation;
     GdkPixbufAnimationIter *iter;
@@ -454,7 +315,6 @@ _gdk_pixbuf_new_from_uri_at_scale (const char *uri,
     GFile *file;
     GFileInfo *file_info;
     GInputStream *input_stream;
-    GError *error = NULL;
 
     g_return_val_if_fail (uri != NULL, NULL);
 
@@ -484,13 +344,20 @@ _gdk_pixbuf_new_from_uri_at_scale (const char *uri,
     }
 
     if (input_stream == NULL) {
-	    input_stream = G_INPUT_STREAM (g_file_read (file, NULL, &error));
-	    if (input_stream == NULL) {
-		    g_warning ("Unable to create an input stream for %s: %s", uri, error->message);
-		    g_clear_error (&error);
-		    g_object_unref (file);
-		    return NULL;
-	    }
+        input_stream = G_INPUT_STREAM (g_file_read (file, NULL, NULL));
+        if (input_stream == NULL) {
+	    g_object_unref (file);
+            return NULL;
+        }
+    }
+
+    loader = gdk_pixbuf_loader_new ();
+    if (1 <= width || 1 <= height) {
+        info.width = width;
+        info.height = height;
+	info.input_width = info.input_height = 0;
+        info.preserve_aspect_ratio = preserve_aspect_ratio;        
+        g_signal_connect (loader, "size-prepared", G_CALLBACK (size_prepared_cb), &info);
     }
 
     has_frame = FALSE;
@@ -502,35 +369,19 @@ _gdk_pixbuf_new_from_uri_at_scale (const char *uri,
 					  buffer,
 					  sizeof (buffer),
 					  NULL,
-					  &error);
-        if (bytes_read == -1) {
-            g_warning ("Error reading from %s: %s", uri, error->message);
-            g_clear_error (&error);
-            break;
-        }
+					  NULL);
+	if (bytes_read == -1) {
+	    break;
+	}
 	result = TRUE;
 	if (bytes_read == 0) {
 	    break;
 	}
 
-        if (loader == NULL) {
-            loader = create_loader (file, buffer, bytes_read);
-            if (1 <= width || 1 <= height) {
-              info.width = width;
-              info.height = height;
-              info.input_width = info.input_height = 0;
-              info.preserve_aspect_ratio = preserve_aspect_ratio;
-              g_signal_connect (loader, "size-prepared", G_CALLBACK (size_prepared_cb), &info);
-            }
-            g_assert (loader != NULL);
-        }
-
 	if (!gdk_pixbuf_loader_write (loader,
 				      (unsigned char *)buffer,
 				      bytes_read,
-				      &error)) {
-            g_warning ("Error creating thumbnail for %s: %s", uri, error->message);
-            g_clear_error (&error);
+				      NULL)) {
 	    result = FALSE;
 	    break;
 	}
@@ -545,18 +396,10 @@ _gdk_pixbuf_new_from_uri_at_scale (const char *uri,
 	}
     }
 
-    if (loader == NULL) {
-        /* This can happen if the above loop was exited due to the
-         * g_input_stream_read() call failing. */
-        result = FALSE;
-    } else if (gdk_pixbuf_loader_close (loader, &error) == FALSE) {
-        g_warning ("Error creating thumbnail for %s: %s", uri, error->message);
-        g_clear_error (&error);
-        result = FALSE;
-    }
+    gdk_pixbuf_loader_close (loader, NULL);
 
     if (!result) {
-        g_clear_object (&loader);
+	g_object_unref (G_OBJECT (loader));
 	g_input_stream_close (input_stream, NULL, NULL);
 	g_object_unref (input_stream);
 	g_object_unref (file);
@@ -608,7 +451,11 @@ gnome_desktop_thumbnail_factory_finalize (GObject *object)
       priv->monitors = NULL;
     }
 
-  g_mutex_clear (&priv->lock);
+  if (priv->lock)
+    {
+      g_mutex_free (priv->lock);
+      priv->lock = NULL;
+    }
 
   if (priv->disabled_types)
     {
@@ -693,7 +540,7 @@ update_or_create_thumbnailer (GnomeDesktopThumbnailFactory *factory,
   Thumbnailer *thumb;
   gboolean found = FALSE;
 
-  g_mutex_lock (&priv->lock);
+  g_mutex_lock (priv->lock);
 
   for (l = priv->thumbnailers; l && !found; l = g_list_next (l))
     {
@@ -721,7 +568,7 @@ update_or_create_thumbnailer (GnomeDesktopThumbnailFactory *factory,
         gnome_desktop_thumbnail_factory_add_thumbnailer (factory, thumb);
     }
 
-  g_mutex_unlock (&priv->lock);
+  g_mutex_unlock (priv->lock);
 }
 
 static void
@@ -732,7 +579,7 @@ remove_thumbnailer (GnomeDesktopThumbnailFactory *factory,
   GList *l;
   Thumbnailer *thumb;
 
-  g_mutex_lock (&priv->lock);
+  g_mutex_lock (priv->lock);
 
   for (l = priv->thumbnailers; l; l = g_list_next (l))
     {
@@ -750,89 +597,7 @@ remove_thumbnailer (GnomeDesktopThumbnailFactory *factory,
         }
     }
 
-  g_mutex_unlock (&priv->lock);
-}
-
-static void
-remove_thumbnailers_for_dir (GnomeDesktopThumbnailFactory *factory,
-                             const gchar                  *thumbnailer_dir,
-                             GFileMonitor                 *monitor)
-{
-  GnomeDesktopThumbnailFactoryPrivate *priv = factory->priv;
-  GList *l;
-  Thumbnailer *thumb;
-
-  g_mutex_lock (&priv->lock);
-
-  /* Remove all the thumbnailers inside this @thumbnailer_dir. */
-  for (l = priv->thumbnailers; l; l = g_list_next (l))
-    {
-      thumb = (Thumbnailer *)l->data;
-
-      if (g_str_has_prefix (thumb->path, thumbnailer_dir) == TRUE)
-        {
-          priv->thumbnailers = g_list_delete_link (priv->thumbnailers, l);
-          g_hash_table_foreach_remove (priv->mime_types_map,
-                                       (GHRFunc)remove_thumbnailer_from_mime_type_map,
-                                       (gpointer)thumb->path);
-          thumbnailer_unref (thumb);
-
-          break;
-        }
-    }
-
-  /* Remove the monitor for @thumbnailer_dir. */
-  priv->monitors = g_list_remove (priv->monitors, monitor);
-  g_signal_handlers_disconnect_by_func (monitor, thumbnailers_directory_changed, factory);
-
-  g_mutex_unlock (&priv->lock);
-}
-
-static void
-gnome_desktop_thumbnail_factory_load_thumbnailers_for_dir (GnomeDesktopThumbnailFactory *factory,
-                                                           const gchar                  *path)
-{
-  GnomeDesktopThumbnailFactoryPrivate *priv = factory->priv;
-  GDir *dir;
-  GFile *dir_file;
-  GFileMonitor *monitor;
-  const gchar *dirent;
-
-  dir = g_dir_open (path, 0, NULL);
-  if (!dir)
-      return;
-
-  /* Monitor dir */
-  dir_file = g_file_new_for_path (path);
-  monitor = g_file_monitor_directory (dir_file,
-                                      G_FILE_MONITOR_NONE,
-                                      NULL, NULL);
-  if (monitor)
-    {
-      g_signal_connect (monitor, "changed",
-                        G_CALLBACK (thumbnailers_directory_changed),
-                        factory);
-      priv->monitors = g_list_prepend (priv->monitors, monitor);
-    }
-  g_object_unref (dir_file);
-
-  while ((dirent = g_dir_read_name (dir)))
-    {
-      Thumbnailer *thumb;
-      gchar       *filename;
-
-      if (!g_str_has_suffix (dirent, THUMBNAILER_EXTENSION))
-          continue;
-
-      filename = g_build_filename (path, dirent, NULL);
-      thumb = thumbnailer_new (filename);
-      g_free (filename);
-
-      if (thumb)
-          gnome_desktop_thumbnail_factory_add_thumbnailer (factory, thumb);
-    }
-
-  g_dir_close (dir);
+  g_mutex_unlock (priv->lock);
 }
 
 static void
@@ -863,19 +628,6 @@ thumbnailers_directory_changed (GFileMonitor                 *monitor,
 
       g_free (path);
       break;
-    case G_FILE_MONITOR_EVENT_UNMOUNTED:
-    case G_FILE_MONITOR_EVENT_MOVED:
-      path = g_file_get_path (file);
-      remove_thumbnailers_for_dir (factory, path, monitor);
-
-      if (event_type == G_FILE_MONITOR_EVENT_MOVED)
-          gnome_desktop_thumbnail_factory_load_thumbnailers_for_dir (factory, path);
-
-      g_free (path);
-      break;
-    case G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT:
-    case G_FILE_MONITOR_EVENT_ATTRIBUTE_CHANGED:
-    case G_FILE_MONITOR_EVENT_PRE_UNMOUNT:
     default:
       break;
     }
@@ -894,7 +646,47 @@ gnome_desktop_thumbnail_factory_load_thumbnailers (GnomeDesktopThumbnailFactory
   dirs = get_thumbnailers_dirs ();
   for (i = 0; dirs[i]; i++)
     {
-      gnome_desktop_thumbnail_factory_load_thumbnailers_for_dir (factory, dirs[i]);
+      const gchar *path = dirs[i];
+      GDir *dir;
+      GFile *dir_file;
+      GFileMonitor *monitor;
+      const gchar *dirent;
+
+      dir = g_dir_open (path, 0, NULL);
+      if (!dir)
+        continue;
+
+      /* Monitor dir */
+      dir_file = g_file_new_for_path (path);
+      monitor = g_file_monitor_directory (dir_file,
+                                          G_FILE_MONITOR_NONE,
+                                          NULL, NULL);
+      if (monitor)
+        {
+          g_signal_connect (monitor, "changed",
+                            G_CALLBACK (thumbnailers_directory_changed),
+                            factory);
+          priv->monitors = g_list_prepend (priv->monitors, monitor);
+        }
+      g_object_unref (dir_file);
+
+      while ((dirent = g_dir_read_name (dir)))
+        {
+          Thumbnailer *thumb;
+          gchar       *filename;
+
+          if (!g_str_has_suffix (dirent, THUMBNAILER_EXTENSION))
+            continue;
+
+          filename = g_build_filename (path, dirent, NULL);
+          thumb = thumbnailer_new (filename);
+          g_free (filename);
+
+          if (thumb)
+            gnome_desktop_thumbnail_factory_add_thumbnailer (factory, thumb);
+        }
+
+      g_dir_close (dir);
     }
 
   priv->loaded = TRUE;
@@ -907,7 +699,7 @@ external_thumbnailers_disabled_all_changed_cb (GSettings                    *set
 {
   GnomeDesktopThumbnailFactoryPrivate *priv = factory->priv;
 
-  g_mutex_lock (&priv->lock);
+  g_mutex_lock (priv->lock);
 
   priv->disabled = g_settings_get_boolean (priv->settings, "disable-all");
   if (priv->disabled)
@@ -921,7 +713,7 @@ external_thumbnailers_disabled_all_changed_cb (GSettings                    *set
       gnome_desktop_thumbnail_factory_load_thumbnailers (factory);
     }
 
-  g_mutex_unlock (&priv->lock);
+  g_mutex_unlock (priv->lock);
 }
 
 static void
@@ -931,15 +723,14 @@ external_thumbnailers_disabled_changed_cb (GSettings                    *setting
 {
   GnomeDesktopThumbnailFactoryPrivate *priv = factory->priv;
 
-  g_mutex_lock (&priv->lock);
+  g_mutex_lock (priv->lock);
 
-  if (!priv->disabled)
-    {
-      g_strfreev (priv->disabled_types);
-      priv->disabled_types = g_settings_get_strv (priv->settings, "disable");
-    }
+  if (priv->disabled)
+    return;
+  g_strfreev (priv->disabled_types);
+  priv->disabled_types = g_settings_get_strv (priv->settings, "disable");
 
-  g_mutex_unlock (&priv->lock);
+  g_mutex_unlock (priv->lock);
 }
 
 static void
@@ -958,7 +749,7 @@ gnome_desktop_thumbnail_factory_init (GnomeDesktopThumbnailFactory *factory)
                                                 (GDestroyNotify)g_free,
                                                 (GDestroyNotify)thumbnailer_unref);
   
-  g_mutex_init (&priv->lock);
+  priv->lock = g_mutex_new ();
 
   priv->settings = g_settings_new ("org.gnome.desktop.thumbnailers");
   priv->disabled = g_settings_get_boolean (priv->settings, "disable-all");
@@ -1011,98 +802,6 @@ gnome_desktop_thumbnail_factory_new (GnomeDesktopThumbnailSize size)
   return factory;
 }
 
-static char *
-thumbnail_filename (const char *uri)
-{
-  GChecksum *checksum;
-  guint8 digest[16];
-  gsize digest_len = sizeof (digest);
-  char *file;
-
-  checksum = g_checksum_new (G_CHECKSUM_MD5);
-  g_checksum_update (checksum, (const guchar *) uri, strlen (uri));
-
-  g_checksum_get_digest (checksum, digest, &digest_len);
-  g_assert (digest_len == 16);
-
-  file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
-
-  g_checksum_free (checksum);
-
-  return file;
-}
-
-static char *
-thumbnail_path (const char                *uri,
-                GnomeDesktopThumbnailSize  size)
-{
-  char *path, *file;
-
-  file = thumbnail_filename (uri);
-  path = g_build_filename (g_get_user_cache_dir (),
-                           "thumbnails",
-                           size == GNOME_DESKTOP_THUMBNAIL_SIZE_LARGE ? "large" : "normal",
-                           file,
-                           NULL);
-  g_free (file);
-  return path;
-}
-
-static char *
-thumbnail_failed_path (const char *uri)
-{
-  char *path, *file;
-
-  file = thumbnail_filename (uri);
-  /* XXX: appname is only used for failed thumbnails. Is this a mistake? */
-  path = g_build_filename (g_get_user_cache_dir (),
-                           "thumbnails",
-                           "fail",
-                           appname,
-                           file,
-                           NULL);
-  g_free (file);
-  return path;
-}
-
-static char *
-validate_thumbnail_path (char                      *path,
-                         const char                *uri,
-                         time_t                     mtime,
-                         GnomeDesktopThumbnailSize  size)
-{
-  GdkPixbuf *pixbuf;
-
-  pixbuf = gdk_pixbuf_new_from_file (path, NULL);
-  if (pixbuf == NULL ||
-      !gnome_desktop_thumbnail_is_valid (pixbuf, uri, mtime)) {
-      g_free (path);
-      return NULL;
-  }
-
-  g_clear_object (&pixbuf);
-
-  return path;
-}
-
-static char *
-lookup_thumbnail_path (const char                *uri,
-                       time_t                     mtime,
-                       GnomeDesktopThumbnailSize  size)
-{
-  char *path = thumbnail_path (uri, size);
-  return validate_thumbnail_path (path, uri, mtime, size);
-}
-
-static char *
-lookup_failed_thumbnail_path (const char                *uri,
-                              time_t                     mtime,
-                              GnomeDesktopThumbnailSize  size)
-{
-  char *path = thumbnail_failed_path (uri);
-  return validate_thumbnail_path (path, uri, mtime, size);
-}
-
 /**
  * gnome_desktop_thumbnail_factory_lookup:
  * @factory: a #GnomeDesktopThumbnailFactory
@@ -1123,10 +822,46 @@ gnome_desktop_thumbnail_factory_lookup (GnomeDesktopThumbnailFactory *factory,
 					time_t                 mtime)
 {
   GnomeDesktopThumbnailFactoryPrivate *priv = factory->priv;
+  char *path, *file;
+  GChecksum *checksum;
+  guint8 digest[16];
+  gsize digest_len = sizeof (digest);
+  GdkPixbuf *pixbuf;
+  gboolean res;
 
   g_return_val_if_fail (uri != NULL, NULL);
 
-  return lookup_thumbnail_path (uri, mtime, priv->size);
+  res = FALSE;
+
+  checksum = g_checksum_new (G_CHECKSUM_MD5);
+  g_checksum_update (checksum, (const guchar *) uri, strlen (uri));
+
+  g_checksum_get_digest (checksum, digest, &digest_len);
+  g_assert (digest_len == 16);
+
+  file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
+  
+  path = g_build_filename (g_get_user_cache_dir (),
+			   "thumbnails",
+			   (priv->size == GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL)?"normal":"large",
+			   file,
+			   NULL);
+  g_free (file);
+
+  pixbuf = gdk_pixbuf_new_from_file (path, NULL);
+  if (pixbuf != NULL)
+    {
+      res = gnome_desktop_thumbnail_is_valid (pixbuf, uri, mtime);
+      g_object_unref (pixbuf);
+    }
+
+  g_checksum_free (checksum);
+
+  if (res)
+    return path;
+
+  g_free (path);
+  return FALSE;
 }
 
 /**
@@ -1150,17 +885,42 @@ gnome_desktop_thumbnail_factory_has_valid_failed_thumbnail (GnomeDesktopThumbnai
 							    const char            *uri,
 							    time_t                 mtime)
 {
-  char *path;
+  char *path, *file;
+  GdkPixbuf *pixbuf;
+  gboolean res;
+  GChecksum *checksum;
+  guint8 digest[16];
+  gsize digest_len = sizeof (digest);
 
-  g_return_val_if_fail (uri != NULL, FALSE);
+  checksum = g_checksum_new (G_CHECKSUM_MD5);
+  g_checksum_update (checksum, (const guchar *) uri, strlen (uri));
 
-  path = lookup_failed_thumbnail_path (uri, mtime, factory->priv->size);
-  if (path == NULL)
-    return FALSE;
+  g_checksum_get_digest (checksum, digest, &digest_len);
+  g_assert (digest_len == 16);
 
+  res = FALSE;
+
+  file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
+
+  path = g_build_filename (g_get_user_cache_dir (),
+			   "thumbnails/fail",
+			   appname,
+			   file,
+			   NULL);
+  g_free (file);
+
+  pixbuf = gdk_pixbuf_new_from_file (path, NULL);
   g_free (path);
 
-  return TRUE;
+  if (pixbuf)
+    {
+      res = gnome_desktop_thumbnail_is_valid (pixbuf, uri, mtime);
+      g_object_unref (pixbuf);
+    }
+
+  g_checksum_free (checksum);
+
+  return res;
 }
 
 static gboolean
@@ -1244,7 +1004,7 @@ gnome_desktop_thumbnail_factory_can_thumbnail (GnomeDesktopThumbnailFactory *fac
   if (!mime_type)
     return FALSE;
 
-  g_mutex_lock (&factory->priv->lock);
+  g_mutex_lock (factory->priv->lock);
   if (!gnome_desktop_thumbnail_factory_is_disabled (factory, mime_type))
     {
       Thumbnailer *thumb;
@@ -1252,7 +1012,7 @@ gnome_desktop_thumbnail_factory_can_thumbnail (GnomeDesktopThumbnailFactory *fac
       thumb = g_hash_table_lookup (factory->priv->mime_types_map, mime_type);
       have_script = thumbnailer_try_exec (thumb);
     }
-  g_mutex_unlock (&factory->priv->lock);
+  g_mutex_unlock (factory->priv->lock);
 
   if (have_script || mimetype_supported_by_gdk_pixbuf (mime_type))
     {
@@ -1375,7 +1135,7 @@ gnome_desktop_thumbnail_factory_generate_thumbnail (GnomeDesktopThumbnailFactory
   pixbuf = NULL;
 
   script = NULL;
-  g_mutex_lock (&factory->priv->lock);
+  g_mutex_lock (factory->priv->lock);
   if (!gnome_desktop_thumbnail_factory_is_disabled (factory, mime_type))
     {
       Thumbnailer *thumb;
@@ -1384,7 +1144,7 @@ gnome_desktop_thumbnail_factory_generate_thumbnail (GnomeDesktopThumbnailFactory
       if (thumb)
         script = g_strdup (thumb->command);
     }
-  g_mutex_unlock (&factory->priv->lock);
+  g_mutex_unlock (factory->priv->lock);
   
   if (script)
     {
@@ -1406,7 +1166,7 @@ gnome_desktop_thumbnail_factory_generate_thumbnail (GnomeDesktopThumbnailFactory
 	    }
 
 	  g_free (expanded_script);
-	  g_unlink (tmpname);
+	  g_unlink(tmpname);
 	  g_free (tmpname);
 	}
 
@@ -1476,88 +1236,89 @@ gnome_desktop_thumbnail_factory_generate_thumbnail (GnomeDesktopThumbnailFactory
 }
 
 static gboolean
-save_thumbnail (GdkPixbuf  *pixbuf,
-                char       *path,
-                const char *uri,
-                time_t      mtime)
+make_thumbnail_dirs (GnomeDesktopThumbnailFactory *factory)
 {
-  char *dirname;
-  char *tmp_path = NULL;
-  int tmp_fd;
-  char mtime_str[21];
-  gboolean ret = FALSE;
-  GError *error = NULL;
-  const char *width, *height;
+  char *thumbnail_dir;
+  char *image_dir;
+  gboolean res;
 
-  if (pixbuf == NULL)
-    return FALSE;
+  res = FALSE;
 
-  dirname = g_path_get_dirname (path);
-
-  if (g_mkdir_with_parents (dirname, 0700) != 0)
-    goto out;
-
-  tmp_path = g_strconcat (path, ".XXXXXX", NULL);
-  tmp_fd = g_mkstemp (tmp_path);
-
-  if (tmp_fd == -1)
-    goto out;
-  close (tmp_fd);
-
-  g_snprintf (mtime_str, 21, "%ld",  mtime);
-  width = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::Image::Width");
-  height = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::Image::Height");
-
-  error = NULL;
-  if (width != NULL && height != NULL)
-    ret = gdk_pixbuf_save (pixbuf,
-			   tmp_path,
-			   "png", &error,
-			   "tEXt::Thumb::Image::Width", width,
-			   "tEXt::Thumb::Image::Height", height,
-			   "tEXt::Thumb::URI", uri,
-			   "tEXt::Thumb::MTime", mtime_str,
-			   "tEXt::Software", "GNOME::ThumbnailFactory",
-			   NULL);
-  else
-    ret = gdk_pixbuf_save (pixbuf,
-			   tmp_path,
-			   "png", &error,
-			   "tEXt::Thumb::URI", uri,
-			   "tEXt::Thumb::MTime", mtime_str,
-			   "tEXt::Software", "GNOME::ThumbnailFactory",
-			   NULL);
-
-  if (!ret)
-    goto out;
-
-  g_chmod (tmp_path, 0600);
-  g_rename (tmp_path, path);
-
- out:
-  if (error != NULL)
+  thumbnail_dir = g_build_filename (g_get_user_cache_dir (),
+				    "thumbnails",
+				    NULL);
+  if (!g_file_test (thumbnail_dir, G_FILE_TEST_IS_DIR))
     {
-      g_warning ("Failed to create thumbnail %s: %s", tmp_path, error->message);
-      g_error_free (error);
+      g_mkdir (thumbnail_dir, 0700);
+      res = TRUE;
     }
-  g_unlink (tmp_path);
-  g_free (tmp_path);
-  g_free (dirname);
-  return ret;
+
+  image_dir = g_build_filename (thumbnail_dir,
+				(factory->priv->size == GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL)?"normal":"large",
+				NULL);
+  if (!g_file_test (image_dir, G_FILE_TEST_IS_DIR))
+    {
+      g_mkdir (image_dir, 0700);
+      res = TRUE;
+    }
+
+  g_free (thumbnail_dir);
+  g_free (image_dir);
+  
+  return res;
 }
 
-static GdkPixbuf *
-make_failed_thumbnail (void)
+static gboolean
+make_thumbnail_fail_dirs (GnomeDesktopThumbnailFactory *factory)
 {
-  return gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 1, 1);
+  char *thumbnail_dir;
+  char *fail_dir;
+  char *app_dir;
+  gboolean res;
+
+  res = FALSE;
+
+  thumbnail_dir = g_build_filename (g_get_user_cache_dir (),
+				    "thumbnails",
+				    NULL);
+  if (!g_file_test (thumbnail_dir, G_FILE_TEST_IS_DIR))
+    {
+      g_mkdir (thumbnail_dir, 0700);
+      res = TRUE;
+    }
+
+  fail_dir = g_build_filename (thumbnail_dir,
+			       "fail",
+			       NULL);
+  if (!g_file_test (fail_dir, G_FILE_TEST_IS_DIR))
+    {
+      g_mkdir (fail_dir, 0700);
+      res = TRUE;
+    }
+
+  app_dir = g_build_filename (fail_dir,
+			      appname,
+			      NULL);
+  if (!g_file_test (app_dir, G_FILE_TEST_IS_DIR))
+    {
+      g_mkdir (app_dir, 0700);
+      res = TRUE;
+    }
+
+  g_free (thumbnail_dir);
+  g_free (fail_dir);
+  g_free (app_dir);
+  
+  return res;
 }
 
+
 /**
  * gnome_desktop_thumbnail_factory_save_thumbnail:
  * @factory: a #GnomeDesktopThumbnailFactory
- * @thumbnail: the thumbnail as a pixbuf
+ * @thumbnail: the thumbnail as a pixbuf 
  * @uri: the uri of a file
- * @original_mtime: the modification time of the original file
+ * @original_mtime: the modification time of the original file 
  *
  * Saves @thumbnail at the right place. If the save fails a
  * failed thumbnail is written.
@@ -1572,18 +1333,91 @@ gnome_desktop_thumbnail_factory_save_thumbnail (GnomeDesktopThumbnailFactory *fa
 						const char            *uri,
 						time_t                 original_mtime)
 {
-  char *path;
+  GnomeDesktopThumbnailFactoryPrivate *priv = factory->priv;
+  char *path, *file;
+  char *tmp_path;
+  const char *width, *height;
+  int tmp_fd;
+  char mtime_str[21];
+  gboolean saved_ok;
+  GChecksum *checksum;
+  guint8 digest[16];
+  gsize digest_len = sizeof (digest);
 
-  path = thumbnail_path (uri, factory->priv->size);
-  if (!save_thumbnail (thumbnail, path, uri, original_mtime))
+  checksum = g_checksum_new (G_CHECKSUM_MD5);
+  g_checksum_update (checksum, (const guchar *) uri, strlen (uri));
+
+  g_checksum_get_digest (checksum, digest, &digest_len);
+  g_assert (digest_len == 16);
+
+  file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
+
+  path = g_build_filename (g_get_user_cache_dir (),
+			   "thumbnails",
+			   (priv->size == GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL)?"normal":"large",
+			   file,
+			   NULL);
+
+  g_free (file);
+
+  g_checksum_free (checksum);
+
+  tmp_path = g_strconcat (path, ".XXXXXX", NULL);
+
+  tmp_fd = g_mkstemp (tmp_path);
+  if (tmp_fd == -1 &&
+      make_thumbnail_dirs (factory))
     {
-      thumbnail = make_failed_thumbnail ();
-      g_free (path);
-      path = thumbnail_failed_path (uri);
-      save_thumbnail (thumbnail, path, uri, original_mtime);
-      g_object_unref (thumbnail);
+      g_free (tmp_path);
+      tmp_path = g_strconcat (path, ".XXXXXX", NULL);
+      tmp_fd = g_mkstemp (tmp_path);
     }
+
+  if (tmp_fd == -1)
+    {
+      gnome_desktop_thumbnail_factory_create_failed_thumbnail (factory, uri, original_mtime);
+      g_free (tmp_path);
+      g_free (path);
+      return;
+    }
+  close (tmp_fd);
+  
+  g_snprintf (mtime_str, 21, "%ld",  original_mtime);
+  width = gdk_pixbuf_get_option (thumbnail, "tEXt::Thumb::Image::Width");
+  height = gdk_pixbuf_get_option (thumbnail, "tEXt::Thumb::Image::Height");
+
+  if (width != NULL && height != NULL) 
+    saved_ok  = gdk_pixbuf_save (thumbnail,
+				 tmp_path,
+				 "png", NULL,
+				 "tEXt::Thumb::Image::Width", width,
+				 "tEXt::Thumb::Image::Height", height,
+				 "tEXt::Thumb::URI", uri,
+				 "tEXt::Thumb::MTime", mtime_str,
+				 "tEXt::Software", "GNOME::ThumbnailFactory",
+				 NULL);
+  else
+    saved_ok  = gdk_pixbuf_save (thumbnail,
+				 tmp_path,
+				 "png", NULL,
+				 "tEXt::Thumb::URI", uri,
+				 "tEXt::Thumb::MTime", mtime_str,
+				 "tEXt::Software", "GNOME::ThumbnailFactory",
+				 NULL);
+    
+
+  if (saved_ok)
+    {
+      g_chmod (tmp_path, 0600);
+      g_rename(tmp_path, path);
+    }
+  else
+    {
+      gnome_desktop_thumbnail_factory_create_failed_thumbnail (factory, uri, original_mtime);
+    }
+
   g_free (path);
+  g_free (tmp_path);
 }
 
 /**
@@ -1604,14 +1438,70 @@ gnome_desktop_thumbnail_factory_create_failed_thumbnail (GnomeDesktopThumbnailFa
 							 const char            *uri,
 							 time_t                 mtime)
 {
-  char *path;
+  char *path, *file;
+  char *tmp_path;
+  int tmp_fd;
+  char mtime_str[21];
+  gboolean saved_ok;
   GdkPixbuf *pixbuf;
+  GChecksum *checksum;
+  guint8 digest[16];
+  gsize digest_len = sizeof (digest);
 
-  path = thumbnail_failed_path (uri);
-  pixbuf = make_failed_thumbnail ();
-  save_thumbnail (pixbuf, path, uri, mtime);
-  g_free (path);
+  checksum = g_checksum_new (G_CHECKSUM_MD5);
+  g_checksum_update (checksum, (const guchar *) uri, strlen (uri));
+
+  g_checksum_get_digest (checksum, digest, &digest_len);
+  g_assert (digest_len == 16);
+
+  file = g_strconcat (g_checksum_get_string (checksum), ".png", NULL);
+
+  path = g_build_filename (g_get_user_cache_dir (),
+			   "thumbnails/fail",
+			   appname,
+			   file,
+			   NULL);
+  g_free (file);
+
+  g_checksum_free (checksum);
+
+  tmp_path = g_strconcat (path, ".XXXXXX", NULL);
+
+  tmp_fd = g_mkstemp (tmp_path);
+  if (tmp_fd == -1 &&
+      make_thumbnail_fail_dirs (factory))
+    {
+      g_free (tmp_path);
+      tmp_path = g_strconcat (path, ".XXXXXX", NULL);
+      tmp_fd = g_mkstemp (tmp_path);
+    }
+
+  if (tmp_fd == -1)
+    {
+      g_free (tmp_path);
+      g_free (path);
+      return;
+    }
+  close (tmp_fd);
+  
+  g_snprintf (mtime_str, 21, "%ld",  mtime);
+  pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 1, 1);
+  saved_ok  = gdk_pixbuf_save (pixbuf,
+			       tmp_path,
+			       "png", NULL, 
+			       "tEXt::Thumb::URI", uri,
+			       "tEXt::Thumb::MTime", mtime_str,
+			       "tEXt::Software", "GNOME::ThumbnailFactory",
+			       NULL);
   g_object_unref (pixbuf);
+  if (saved_ok)
+    {
+      g_chmod (tmp_path, 0600);
+      g_rename(tmp_path, path);
+    }
+
+  g_free (path);
+  g_free (tmp_path);
 }
 
 /**
@@ -1649,7 +1539,23 @@ char *
 gnome_desktop_thumbnail_path_for_uri (const char         *uri,
 				      GnomeDesktopThumbnailSize  size)
 {
-  return thumbnail_path (uri, size);
+  char *md5;
+  char *file;
+  char *path;
+
+  md5 = gnome_desktop_thumbnail_md5 (uri);
+  file = g_strconcat (md5, ".png", NULL);
+  g_free (md5);
+  
+  path = g_build_filename (g_get_user_cache_dir (),
+			   "thumbnails",
+			   (size == GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL)?"normal":"large",
+			   file,
+			   NULL);
+    
+  g_free (file);
+
+  return path;
 }
 
 /**
